;;;; -*- mode:Scheme -*- ;;;;
;;
;; loader.scm
;; 
;; This is the file you actually want to execute.
;; It loads all five code files, then the knowledge file, 
;; then executes a best-first search for the desired solution.
;;

;; Load the files (compiled or otherwise):
(cd "/mit/6.034/Project/")
(load "search")
(load "rule-utils")
(load "forward")
(load "match")
(load "schedule")
(cd "/mit/catfoo/6.034")



(define (extend-path-with-heuristic path heuristic)
;  (display* "Extending the path " (path-reverse path))
  (display* (last (path-head-node path)))
  (remove-falses
   (map
    (lambda (next-node)
      (if (member next-node (path-nodes path))
	  #f
	  (make-path
	   (heuristic next-node)	; path value is from calling heuristic
	   (cons next-node (path-nodes path)))))
    (get-neighboring-nodes (path-head-node path)))))

(define (last node)
	(if (null? (cdr node))
		node
		(last (cdr node))))


(define (get-heuristic-value node finish)
  (define (loop assertions score)
    (cond ((null? assertions) score)
	  ((member (car assertions) node)
	   (loop (cdr assertions) (- score 1)))
	  (else
	   (loop (cdr assertions) score))))
  (fluid-let ((*assertions* (copy-list node)))  
    (+ (loop finish 0)
       (* -1 (length (build-rule-instances '(r1 if ((? ta) is assigned to (? course))
                                                   ((? ta) prefers (? course))
                                               add ((? ta) has preference for (? course) satisfied)))))
       (* -1 (length (build-rule-instances '(r2 if ((? ta) is assigned to (? course))
                                                   ((? course) prefers TA who prefers (? course))
                                                   ((? ta) prefers (? course))
                                               add ((? ta) has preference for (? course) satisfied)))))
       (* -1 (length (build-rule-instances '(r3 if ((? ta) is assigned to (? course))
                                                   ((? course) prefers TA who is (? adj))
                                                   ((? ta) is (? adj))
                                               add ((? ta) has preference for (? course) satisfied))))))))




(define (beam k start finish)

  (define *new-paths* '())

  (define (merge-paths-into-Q new-paths Q)
  
    (define (first-n-paths n list new-list)
      (if (or (= n 0) (null? list))
          new-list
          (first-n-paths (- n 1) (cdr list) (append new-list (car list)))))
    (begin (pp Q)
	   (pp new-paths)
	   (pp *new-paths*)
	   (cond ((Q-empty? Q)
		  (Q-set-paths! Q (first-n-paths k
						 (sort-paths-by-value (append new-paths *new-paths*))
						 '()))
		  (set! *new-paths* '()))
		 (else
		  (set! *new-paths* (append new-paths *new-paths*))))
	   Q))

  (define (heuristic node) 
    ;; in general, the heuristic value may depend on the finishn node
    (get-heuristic-value node finish))

  (define (successors path)
    (extend-path-with-heuristic path heuristic))

  (set! *number-of-search-steps* 0)

  ;; Fire up generalized search using Q constructor defined above:
  (search
   ;; Just start with a partial path including only start node.
   finish				; goal node
   1					; only 1 path wanted
   pick-and-remove-best-path		; pick the best
   merge-paths-into-Q			; add to the front of Q
   successors				; successors, using heuristic
   ;; The initial Q, just one path = (start)
   (make-Q (make-path (heuristic start) (list start))) ; initial Q
   '()					; initial wins
   ))


(read-k-file "rules.txt")

;(display-assertions)

(beam 5 *assertions* '(
;(6AAA has TA) (6BBB has TA) (6CCC has TA) (6DDD has TA)
;                           (6LLL has TA) (6MMM has TA) (6NNN has TA) (6034 has TA)
                          (6EEE has TA) (6III has TA)
))

(get-step-count)

