package ex2;

import java.util.StringTokenizer;

/**
   PERFORMANCE
   this code fails the following tests:
   5) testDiv(ex2.RatPolyTest) "expected:<x^7-x^6+x^5-x^4+x^3-x^2+x-1> but was:<x^8>"
   6) testDivComplexI(ex2.RatPolyTest) "expected:<x+1> but was:<1>"
   7) testDivComplexII(ex2.RatPolyTest) "expected:<x^3+2*x+1> but was:<1>"
   8) testDivExamplesFromSpec(ex2.RatPolyTest) "expected:<1/3*x> but was:<0>"
   9) testDivExampleFromPset(ex2.RatPolyTest) "expected:<1/3*x^2-2/9> but was:<1>"

   this code seems to work properly on the following tests when tried outside
   of RatPolyTest - however, it creates an AssertionFailedError.
   1) testParseMultTerms(ex2.RatPolyTest)
   2) testParseLeadingNeg(ex2.RatPolyTest)
   3) testParseLeadingConstants(ex2.RatPolyTest)
   4) testParseRationals(ex2.RatPolyTest)
   
**/

/** RatPoly represents an immutable single-variate polynomial
    expression.  RatPolys have RatNum coefficients and integer
    exponents.
    <p>
    
    Examples of RatPolys include "0", "x-10", and "x^3-2*x^2+5/3*x+3",
    and "NaN".  
*/
public class RatPoly {

    // holds terms of this
    private RatTermVec terms;

    private static RatNum zero = new RatNum(0);
    private static RatNum one = new RatNum(1);

    // Definitions:
    //   For a RatPoly p, let C(p,i) be "p.terms.get(i).coeff" and
    //                        E(p,i) be "p.terms.get(i).expt"
    //                        length(p) be "p.terms.size()"
    // (These are helper functions that will make it easier for us
    // to write the remainder of the specifications.  They are not
    // executable code; they just represent complex expressions in a
    // concise manner, so that we can stress the important parts of
    // other expressions in the spec rather than get bogged down in
    // the details of how we extract the coefficient for the 2nd term
    // or the exponent for the 5th term.  So when you see C(p,i),
    // think "coefficient for the ith term in p".)

    // Abstraction Function:
    //   A RatPoly p is the Sum, from i=0 to length(p), of C(p,i)*x^E(p,i)
    // (This explains what the state of the fields in a RatPoly
    // represents: it is the sum of a series of terms, forming an
    // expression like "C_0 + C_1*x^1 + C_2*x^2 + ...".  If there are no
    // terms, then the RatPoly represents the zero polynomial.)

    // Representation Invariant for every RatPoly p:
    //   terms != null &&
    //   forall i such that (0 <= i < length(p)), C(p,i) != 0 &&
    //   forall i such that (0 <= i < length(p)), E(p,i) >= 0 &&
    //   forall i such that (0 <= i < length(p) - 1), E(p,i) > E(p, i+1)
    // (This tells us four important facts about every RatPoly:
    //   * the terms field always points to some usable object,
    //   * no term in a RatPoly has a zero coefficient,
    //   * no term in a RatPoly has a negative exponent, and
    //   * the terms in a RatPoly are sorted in descending exponent order.)

    /** @effects: Constructs a new Poly, "0".
     */
  public RatPoly() {
    terms = new RatTermVec();
  }

    /** @requires: e >= 0
	@effects: Constructs a new Poly equal to "c * x^e".
	If c is zero, constructs a "0" polynomial.
     */
  public RatPoly(int c, int e) {
    RatNum newRat;
    RatTerm newterm;
    terms = new RatTermVec();
    if (e >= 0) {
      if (c != 0) {
	newRat = new RatNum(c);
	newterm = new RatTerm(newRat, e);
	terms.addElement(newterm);
      }
    }
  }
    /** @requires: 'rt' satisfies clauses given in rep. invariant
	@effects: Constructs a new Poly using 'rt' (does not make a copy)
    */
  private RatPoly(RatTermVec rt) {
    terms = rt;
  }


    /** Returns the degree of this.
	@requires: !this.isNaN()
	@return the largest exponent with a non-zero coefficient, or 0
	        if this is "0".
    */
  public int degree() {
    int size = terms.size();
    if (size == 0)
      return 0;
    else 
      return (terms.get(size - 1)).expt;
  }
  
    /** @requires: !this.isNaN()
	@return the coefficient associated with term of degree 'deg'.
        If there is no term of degree 'deg' in this poly, then returns
        zero.
    */
  public RatNum coeff(int deg) {
    int i = findTermIndex(terms, deg);
    RatTerm hg = hintedGet(terms, i, deg);
    return hg.coeff;
  }

    /** @return true if and only if this has some coefficient = "NaN".
     */
  public boolean isNaN() {
    int size = terms.size();
    for (int i = 0; i < size; i++) {
      if (((terms.get(i)).coeff).isNaN())
	return true;
    }
    return false;
  }
  
    /** Returns a string representation of this.
	@return a String representation of the expression represented
	by this, with the terms sorted in order of degree from highest
	to lowest.
	<p>
	There is no whitespace in the returned string.
	<p>
	Terms with zero coefficients do not appear in the returned
	string, unless the polynomial is itself zero, in which case
	the returned string will just be "0".
	<p>
	If this.isNaN(), then the returned string will be just "NaN"
	<p>
	The string for a non-zero, non-NaN poly is in the form 
	"(-)T(+|-)T(+|-)...",where for each
	term, T takes the form "C*x^E" or "C*x", UNLESS: 
	(1) the exponent E is zero, in which case T takes the form "C", or 
	(2) the coefficient C is one, in which case T takes the form 
	    "x^E" or "x"
	<p>
	Note that this format allows only the first term to be output
	as a negative expression.
	<p>
	Valid example outputs include "x^17-3/2*x^2+1", "-x+1", "-1/2",
	and "0".
	<p>
    */
    public String unparse()
    {
	// IMPL IS STAFF ONLY
	checkRep();
	if (terms.size() == 0) {
	    checkRep();
	    return "0";
	}

	StringBuffer sb = new StringBuffer();

	// first term, special case to omit initial '+'
	RatTerm term = terms.get(0);
	if (term.coeff.isNegative()) {
	    sb.append("-");
	    term = new RatTerm( term.coeff.negate(), term.expt );
	}
	appendTerm(sb, term);
	
	for(int i=1; i<terms.size(); i++) {
	    RatTerm rt = terms.get(i);
	    RatNum num = rt.coeff;
	    if (num.equals(zero)) {
		continue; // do nothing;
	    } else if (num.isNegative()) {
		sb.append("-");
		num = num.negate();
	    } else {
		sb.append("+");
	    }
	    appendTerm(sb, new RatTerm( num, rt.expt ));
	}
	
	if (sb.length() == 0) 
	    sb.append('0');

	checkRep();
	return sb.toString(); 
    }

    // METHOD IS STAFF ONLY
    // appendTerm is a helper method, not given in spec
    // Let c = rt.coeff and e = rt.expt
    // If c is zero then does nothing.  
    // Else, does the following in series:
    //    1) if c != 1 or e = 0 then appends c to sb else does nothing
    //    2) if c != 1 and e != 0 then appends "*" to sb
    //    3) if e > 1 then appends ("x^"+e) to sb.
    //    4) if e == 1 then appends ("x") to sb.
    private static void appendTerm(StringBuffer sb, RatTerm rt) {
	RatNum c = rt.coeff;
	int e = rt.expt;

	if(sb.toString().equals("NaN"))
        {
            return;
        }
           
        if(c.isNaN())
        {
            sb.replace(0, sb.length(), "NaN");
            return;
        }
	
	if (c.equals( zero ) ) {
	    return;
	}
	
	if (e == 0 || !c.equals( one ) ) {
	    sb.append(c.unparse());
	}

	if (e != 0) {
	    if (!c.equals( one )) {
		sb.append('*');		
	    }
	    sb.append('x');
	    if (e != 1) {
		sb.append('^');
		sb.append(e);
	    } 
	}
    }
        

    /** Scales coefficients within 'vec' by 'scalar' (helper procedure).
	@requires: vec, scalar != null
	@modifies: vec
	@effects: Forall i s.t. 0 <= i < vec.size(),
	             if (C . E) = vec.get(i)
		     then vec_post.get(i) = (C*scalar . E)
	@see ex2.RatTerm regarding (C . E) notation
    */
  private static void scaleCoeff(RatTermVec vec, RatNum scalar) {
    RatTerm curTerm;
    RatNum newC;
    int size = vec.size();
    for (int i = 0; i < size; i++) {
      curTerm = vec.get(i);
      newC = (curTerm.coeff).mul(scalar);
      vec.set(new RatTerm(newC, curTerm.expt), i);
    }
  }

    /** Increments exponents within 'vec' by 'degree' (helper procedure).
	@requires: vec != null
	@modifies: vec
	@effects: Forall i s.t. 0 <= i < vec.size(),
	             if (C . E) = vec.get(i)
		     then vec_post.get(i) = (C . E+degree)
	@see ex2.RatTerm regarding (C . E) notation
     */
  private static void incremExpt(RatTermVec vec, int degree) {
    RatTerm curTerm;
    int newE;
    int size = vec.size();
    for (int i = 0; i < size; i++) {
      curTerm = vec.get(i);
      newE = curTerm.expt * degree;
      vec.set(new RatTerm(curTerm.coeff, newE), i);
    }
  }
    /** Merges a term into a sequence of terms, preserving the
	sorted nature of the sequence (helper procedure).

	Definitions:
	Let a "Sorted RatTermVec" be a RatTermVec V such that
	[1] V is sorted in descending exponent order &&
	[2] there are no two RatTerms with the same exponent in V &&
	[3] there is no RatTerm in V with a coefficient equal to zero

	For a Sorted(RatTermVec) V and integer e, let cofind(V, e)
	be either the coefficient for a RatTerm rt in V whose
	exponent is e, or zero if there does not exist any such
	RatTerm in V.  (This is like the coeff function of RatPoly.)

	@requires: vec != null && sorted(vec)
	@modifies: vec
	@effects: sorted(vec_post) &&
	          cofind(vec_post,newTerm.expt)
                    = cofind(vec,newTerm.expt) + newTerm.coeff
    */
  private static void sortedAdd(RatTermVec vec, RatTerm newTerm) {
    RatTerm curTerm, anotherTerm;
    int E = newTerm.expt;
    
    int possI = findTermIndex(vec, E);
    curTerm = hintedGet(vec, possI, E);
    anotherTerm = new RatTerm(curTerm.coeff.add(newTerm.coeff), E);

    if (curTerm.coeff.equals(zero)) 
      vec.insert(anotherTerm, possI);
    else {
      if (anotherTerm.coeff.equals(zero))
	vec.remove(possI);
      else
	vec.set(anotherTerm, possI);
    }
      
  }

    /** @return a new Poly equal to "0 - this";
	        if this.isNaN(), returns some r such that r.isNaN()
     */
  public RatPoly negate() {
    RatTerm tm, newrt;
    RatTermVec rtv = new RatTermVec();
    int size = terms.size();

    checkRep();
    for (int i = 0; i < size; i++) {
      tm = terms.get(i);
      newrt = new RatTerm(tm.coeff.negate(), tm.expt);
      sortedAdd(rtv,newrt);
    }
    return new RatPoly(rtv);
  }

    /** Addition operation.
	@requires: p != null
	@return a new RatPoly, r, such that r = "this + p";
	        if this.isNaN() or p.isNaN(), returns some r such that r.isNaN()
    */
  public RatPoly add(RatPoly p) {

    int size = terms.size();
    int psize = p.terms.size();
    RatTermVec rtv = new RatTermVec();
    checkRep();
    
    for (int i = 0; i < size; i ++)
      sortedAdd(rtv, terms.get(i));

    for (int i = 0; i < psize; i ++)
      sortedAdd(rtv, p.terms.get(i));

    return new RatPoly(rtv);
  }

    /** Subtraction operation.
	@requires: p != null
	@return a new RatPoly, r, such that r = "this - p";
	        if this.isNaN() or p.isNaN(), returns some r such that r.isNaN()
    */
  public RatPoly sub(RatPoly p) {
    checkRep();
    p.checkRep();
    return this.add(p.negate());
  }

    /** Multiplication operation.
	@requires: p != null
	@return a new RatPoly, r, such that r = "this * p";
	        if this.isNaN() or p.isNaN(), returns some r such that r.isNaN()
    */
  ///// !!!!!!
  public RatPoly mul(RatPoly p) {
    /*    RatTermVec rtv;
    RatPoly RPtotal = new RatPoly();
    int psize = p.terms.size();
    
    for(int i = 0; i < psize; i++) {
      rtv = terms.copy();
      scaleCoeff(rtv, p.terms.get(i).coeff);
      incremExpt(rtv, p.terms.get(i).expt);
      RPtotal = RPtotal.add(new RatPoly(rtv));
    }

    return RPtotal;
    */
    RatNum newRN;
    int newExpt;
    RatTermVec rtv = new RatTermVec();
    checkRep();
    p.checkRep();
    
    int size = terms.size();
    int psize = p.terms.size();

    if ((size == 0 && p.isNaN()) || (psize == 0 && this.isNaN()))
      return RatPoly.parse("NaN");
    
    for (int i = 0; i < size; i++) {
      for (int j = 0; j < psize; j++) {
	newRN = terms.get(i).coeff.mul(p.terms.get(j).coeff);
	newExpt = terms.get(i).expt + p.terms.get(j).expt;
	sortedAdd(rtv, new RatTerm(newRN, newExpt));
      }
    }

    
	return new RatPoly(rtv);
   

  }

    /** Division operation (truncating).
	@requires: p != null
	@return a new RatPoly, q, such that q = "this / p"; 
	        if p = 0 or this.isNaN() or p.isNaN(), 
		returns some q such that q.isNaN()
	<p>

	Division of polynomials is defined as follows:
	Given two polynomials u and v, with v != "0", we can divide u by
	v to obtain a quotient polynomial q and a remainder polynomial
	r satisfying the condition u = "q * v + r", where
          the degree of r is strictly less than the degree of v,
          the degree of q is no greater than the degree of u, and
          r and q have no negative exponents.
	<p>

	For the purposes of this class, the operation "u / v" returns
	q as defined above.
	<p>

	Thus, "x^3-2*x+3" / "3*x^2" = "1/3*x" (with the corresponding
	r = "-2*x+3"), and "x^2+2*x+15 / 2*x^3" = "0" (with the
	corresponding r = "x^2+2*x+15").
	<p>

	Note that this truncating behavior is similar to the behavior
	of integer division on computers.
    */
    public RatPoly div(RatPoly p)
    {
	//STAFF IMPL
	checkRep();
	if (this.isNaN() || p.isNaN() || 
	    p.unparse().equals((new RatPoly()).unparse())) {
	    checkRep();
	    return RatPoly.parse("NaN");
	} else {
	    checkRep();
	    return divAndRem(p)[0];
	}
    }
    

    
    /** @returns a RatPoly[] ps = { "this/p", "this mod p" }
     */
    private RatPoly[] divAndRem(RatPoly p) {
	// Taken from Knuth, "The Art of Computer Programming"
	// sec. 4.6.1, Algorithm D
	checkRep();
	RatPoly u = this, v = p;
	final int m = u.degree();
	final int n = v.degree();
	RatTermVec q_Terms = new RatTermVec();
	RatTermVec r_Terms = u.terms.copy();
	
	// D1: do step D2 for k = m-n, m-n-1, ..., 0; then terminate
	// with (r[n-1], ..., r[0]) = (u[n-1], ..., u[0])
	for(int k=m - n; k >= 0; k--) {  
	    // D2: Set q[k] := u[n+k]/v[n] ...	    
	    RatTerm u_nk = hintedGet(r_Terms, findTermIndex(r_Terms, n+k), n+k);
	    RatNum q_k = u_nk.coeff.div(v.coeff(n));
	    if (!q_k.equals(zero)) {
		q_Terms.addElement(new RatTerm( q_k, k ));
	    }

	    //     Then set u[j] := u[j] - q[k]*v[j-k] 
	    //                      for j=n+k-l, n+k-2, ... k
	    for(int j=n+k-1; j>=k; j--) {
		int u_j_ind = findTermIndex(r_Terms, j);
		RatTerm u_j = hintedGet(r_Terms, u_j_ind, j);
		RatNum v_jk = v.coeff(j-k);

		RatNum new_u_j = u_j.coeff.sub( q_k.mul( v_jk ));
		
		u_j = new RatTerm(new_u_j, j);
		replaceExpt(u_j, r_Terms, u_j_ind);
	    }
	}

	// Repeat: 
	// with (r[n-1], ..., r[0]) = (u[n-1], ..., u[0])
	// (need to fixup r_Terms so that it has no terms from n and up
	while(r_Terms.size() > 0) { // not the only condition; see below
	    RatTerm t = r_Terms.get(0);
	    if (t.expt >= n) {
		r_Terms.remove(0);
	    } else {
		break;
	    }
	}
	
	RatPoly quotient = new RatPoly(q_Terms);
	RatPoly remainder = new RatPoly(r_Terms);

	if (!quotient.mul(p).add(remainder).unparse().equals(this.unparse())) {
	    
	    // check for NaN before we throw a spurious exception... 
	    if (containsNaN(q_Terms) ||
		containsNaN(r_Terms) )
		throw new RuntimeException("invalid q/r pair\n"+
					   "q:"+quotient.terms.printDebug()+"\n"+
					   "r:"+remainder.terms.printDebug()+"\n"+
					   "u:"+this.terms.printDebug()+"\n"+
					   "v:"+p.terms.printDebug()
					   );
	}
	if (remainder.degree() >= p.degree() && p.degree() > 0) {
	    throw new RuntimeException("invalid remainder "+remainder.unparse());
	}
	
	if (quotient.isNaN()) 
	    throw new RuntimeException("has nan...dunno how...");

	checkRep();
	return new RatPoly[] { quotient, remainder };
    }
    
    //Private helper method not in the spec
    private static boolean containsNaN(RatTermVec vec) {
	for(int i=0; i<vec.size(); i++) {
	    RatTerm term = vec.get(i);
	    if (term.coeff.equals(RatNum.parse("NaN"))) {
		return true;
	    }
	}
	return false;
    }

    //Private helper method not in the spec
    private static int findTermIndex(RatTermVec ts, int expt) {
	for(int i=0; i<ts.size(); i++) {
	    RatTerm t = ts.get(i);
	    if (t.expt == expt) {
		return i;
	    } else if (t.expt < expt) {
		return i;
	    }
	}
	return ts.size();
    }

    //Private helper method not in the spec
    private static RatTerm hintedGet(RatTermVec ts, int i, int expt) {
	if (i < ts.size() &&
	    ts.get(i).expt == expt) {
	    return ts.get(i);
	} else {
	    return new RatTerm(zero, expt);
	}
    }

    //Private helper method not in the spec
    private static void replaceExpt(RatTerm term, RatTermVec ts, int index) {
	if (index < ts.size()) {
	    RatTerm t = ts.get(index);
	    if (t.expt == term.expt) {
		if (!term.coeff.equals(zero)) {
		    ts.set(term, index);
		} else {
		    ts.remove(index);
		}
	    } else {
		if (!term.coeff.equals(zero))
		    ts.insert(term, index);
	    }
	} else {
	    if (!term.coeff.equals(zero))
		ts.addElement(term);
	}
    }


    
    /** @return the value of this polynomial when evaluated at 'd'.
	        For example, "x+2" evaluated at 3 is 5, and "x^2-x"
		evaluated at 3 is 6.  Should always return
		Double.NaN if this.isNaN() == true.
     */

  public double eval(double d) {
    int size = terms.size();
    double total = 0;
    double thisterm = 0;
    
    for (int i = 0; i < size; i++) {
      thisterm = 1;
      for (int j = 0; j < terms.get(i).expt; j++) {
	thisterm = thisterm * d;
      }
      total = total + thisterm * terms.get(i).coeff.approx();
    }

    return total;
  }



    /** @requires: 'polyStr' is an instance of a string with no spaces
	           that expresses a poly in the form defined in the
		   unparse() method.
	@return a RatPoly p such that p.unparse() = polyStr
    */
    public static RatPoly parse(String polyStr) {
	RatPoly result = new RatPoly();

	// First we decompose the polyStr into its component terms;
	// third arg orders "+" and "-" to be returned as tokens. 
	StringTokenizer termStrings = 
	    new StringTokenizer(polyStr, "+-", true);

	boolean nextTermIsNegative = false;
	while (termStrings.hasMoreTokens()) {
	    String termToken = termStrings.nextToken();

	    if (termToken.equals("-")) { 
		nextTermIsNegative = true; 
	    } else if (termToken.equals("+")) { 
		nextTermIsNegative = false; 
	    } else {
		// Not "+" or "-"; must be a term

		// Term is: "R" or "R*x" or "R*x^N" or "x^N" or "x",
		// where R is a rational num and N is a natural num. 

		// Decompose the term into its component parts.
		// third arg orders '*' and '^' to act purely as delimiters.
		StringTokenizer numberStrings = 
		    new StringTokenizer(termToken, "*^", false);

		RatNum coeff;
		int expt;
		
		String c1 = numberStrings.nextToken();
		if (c1.equals("x")) { 
		    // ==> "x" or "x^N"
		    coeff = new RatNum(1);

		    if (numberStrings.hasMoreTokens()) {
			// ==> "x^N"
			String N = numberStrings.nextToken();
			expt = Integer.parseInt(N);
			
		    } else {
			// ==> "x"
			expt = 1;
		    }
		} else {
		    // ==> "R" or "R*x" or "R*x^N"
		    String R = c1;
		    coeff = RatNum.parse(R);

		    if (numberStrings.hasMoreTokens()) { 
			// ==> "R*x" or "R*x^N"
			String x = numberStrings.nextToken();

			if (numberStrings.hasMoreTokens()) { 
			    // ==> "R*x^N"
			    String N = numberStrings.nextToken();
			    expt = Integer.parseInt(N);
			} else { 
			    // ==> "R*x"
			    expt = 1;
			}
			
		    } else { 
			// ==> "R"
			expt = 0;
		    }
		}
		
		// at this point, coeff and expt are initialized.
		// Need to fix coeff if it was preceeded by a '-'
		if (nextTermIsNegative) {
		    coeff = coeff.negate();
		}


		// accumulate terms of polynomial in 'result'
		if (!coeff.equals(new RatNum(0))) {
		  RatPoly termPoly = new RatPoly();
		  termPoly.terms.addElement(new RatTerm(coeff, expt));
		  result = result.add(termPoly);
		}
	    }
	}
	return result;
    }


    

    public String toString() {
	checkRep();
	return this.unparse();
    }
    

    private void checkRep() {
	if (terms == null) {
	    throw new RuntimeException("terms == null!");
	}
	for (int i=0; i < terms.size(); i++) {
	    if (terms.get(i).coeff.equals(zero)) {
		throw new RuntimeException("zero coefficient!");
	    }
	    if (terms.get(i).expt < 0) {
		throw new RuntimeException("negative exponent!");
	    }
	    if (i < terms.size() - 1) {
		if (terms.get(i+1).expt >= terms.get(i).expt) {
		    throw new RuntimeException("terms out of order!");
		}
	    }
	}
    }
}

