package ex4;

import java.util.*;


public class Graph {

  HashMap NodeMap;

  
  public Graph() {
    NodeMap = new HashMap();
  }

  // Representation Invariant
  //
  //  A graph is a set of nodes, no duplicates.
  //
  //  Each edge has exactly one label, one origin, and one destination.
  //  (in graph)
  //
  //  A graph associates one node with a list of edges. These edges
  //  all have their origin field pointing to the node.

  void repCheck() {
    
    Object curnode;
    EdgeSet curedges;
    Edge thisedge;
    
    // check that all the edges actually originate from the
    // node they are associated with, and that they point to
    // a node that exists.
    
    for(Iterator allNodes = getNodes().iterator(); allNodes.hasNext(); ) {
      curnode = allNodes.next();
      curedges = getEdges(curnode);
      for(Iterator alledges = curedges.iterator(); alledges.hasNext(); ) {
	thisedge = (Edge) alledges.next();
	if (!(thisedge.origin()).equals(curnode))
	  throw new RuntimeException("edge origin error!");
	else if (! containsNode(thisedge.dest()))
	  throw new RuntimeException("edge dest error!");
	
	// check that each edge has an origin, destination, and label.
	
	else if (thisedge.origin() == null)
	  throw new RuntimeException("edge origin is null!");
	else if (thisedge.dest() == null)
	  throw new RuntimeException("edge dest is null!");
	else if (thisedge.label() == null)
	  throw new RuntimeException("edge label is null!");
      }
    }
    
  }

  // Abstraction Function
  //
  //  graph = nodes + edges
  //  set of nodes = Graph.getNodes();
  //  set of edges = union of all sets from Graph.getEdges(node);

  public String toString() { 
    String graphstring = null;
    
    Object curnode;
    EdgeSet curedges;
    Edge thisedge;
    
    for(Iterator allNodes = getNodes().iterator(); allNodes.hasNext(); ) {
      
      curnode = allNodes.next();
      graphstring = graphstring + curnode + ":\n";
      
      curedges = getEdges(curnode);
      
      for(Iterator alledges = curedges.iterator(); alledges.hasNext(); ) {
	thisedge = (Edge) alledges.next();
	graphstring = graphstring + "\t" + thisedge.label() + " --> " + thisedge.dest + "\n";
      }
    }

    return graphstring;
  }

  
  // adds a node to the current graph.
  // @requires: Node o does not already exist
  // @return: true if the node is successfully added, false if it's
  // already exists
  // @modifies this
  public boolean addNode(Object o) {

    repCheck();
    
    if (NodeMap.containsKey(o)) {
      return false;
    }
    else {

      // inserts the new node with no edges.
      NodeMap.put(o, new EdgeSet());
      repCheck();
      return true;
    }
  }


  // adds an edge
  // @return false if the origin of the edge doesn't exist
  // or if the edge is already there.
  public boolean addEdge(Edge e) {
    EdgeSet oldset, newset;

    repCheck();
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	return false;
      }
      else {
      oldset.add(e);
      NodeMap.put(e.origin(), oldset);

      repCheck();
      return true;
      }
    }
    else return false;
  }


    // adds an edge
  // @return false if the origin of the edge doesn't exist
  // or if the edge is already there.
  public boolean addEdge(String label, Object origin, Object dest) {
    EdgeSet oldset, newset;
    Edge e = new Edge(label, origin, dest);
    
    repCheck();
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	return false;
      }
      else {
      oldset.add(e);
      NodeMap.put(e.origin(), oldset);

      repCheck();
      return true;
      }
    }
    else return false;
  }

  // gets all the nodes from the graph. modifications to the set view
  // will be reflected in the graph.
  // @return set view of nodes in graph.
  public Set getNodes() {
    return NodeMap.keySet();
  }

  public boolean containsNode(Object n) {
    return NodeMap.containsKey(n);
  }
  
  // @return set view of edges in Graph, null if node doesn't exist.   
  public EdgeSet getEdges(Object n) {
    if (this.containsNode(n)) {
      return (EdgeSet) NodeMap.get(n);
    }
    else return null;
  }
  
  public int nodeSize() {
    return NodeMap.size();
  }


  // iterate over all the nodes to count the edges per node.
  public int edgeSize() {
    Object curnode;
    Set keys = this.getNodes();
    int count = 0;
    
    for (Iterator it = keys.iterator(); it.hasNext(); ) {
      curnode = it.next();
      count += this.getEdges(curnode).size();
    }
    return count;
  }

  // MUTATOR, has repcheck()
  // removes node n and all of the edges that come from it.
  public boolean removeNode(Object n) {
    Set keys;
    EdgeSet oldset, newset = new EdgeSet();
    Object curnode;
    Edge thisedge;

    // make sure we have a valid graph object at beginning.
    repCheck();
    
    if (NodeMap.containsKey(n)) {
      NodeMap.remove(n);
      keys = this.getNodes();

      for (Iterator it = keys.iterator(); it.hasNext(); ) {
	curnode = it.next();
	oldset = (EdgeSet) NodeMap.get(curnode);
	newset.removeAll(newset);
	for (Iterator edgeit = oldset.iterator(); edgeit.hasNext(); ) {
	  thisedge = (Edge) edgeit.next();
	  if (thisedge.dest() != n) {
	    newset.add(thisedge);
	  }
	}
	NodeMap.put(curnode, newset);
      }
      repCheck();
      return true;
    }
    else return false;
   
  }

  // returns false if: edge does not exist.
  public boolean removeEdge(Edge e) {
    EdgeSet oldset, newset;

    repCheck();
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	oldset.remove(e);
	NodeMap.put(e.origin(), oldset);
	repCheck();
	
	return true;
      }
    }
    return false;
  }
  

  // returns false if: edge does not exist.
  public boolean removeEdge(String label, Object origin, Object dest) {
    EdgeSet oldset, newset;


    Edge e = new Edge(label, origin, dest);
    repCheck();
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	oldset.remove(e);
	NodeMap.put(e.origin(), oldset);
	repCheck();
	
	return true;
      }
    }
    return false;
  }
  
}



