import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;
import Vertex3D;
import Triangle;
import Matrix4x4;
import MatrixStack;
import Light;
import Surface;
import Luxo;
import Box; 
import Ball;

/*
    LuxoApplet provides only user interface elements.
    All od the interesting stuff is done in the
    LuxoCanvas Class.
*/
public class LuxoApplet extends Applet {
  Panel viewerPanel, buttonPanel;
  LuxoCanvas luxoViewer;
  Button button[];

  // These are the button labels. The index into this array
  // determines the number passed to LuxoCanvas.processButton()
  String bLabel[ ] = {
    "Reset",
    "Sky Cam",
    "Mom Cam",
    "Jr Cam",
    "Mom @ Jr",
    "Mom @ Ball",
    "Jr @ Mom",
    "Jr @ Ball",
    "Move Jr N",
    "Move Jr S",
    "Move Jr E",
    "Move Jr W",
    "Mom Nod",
    "Ball Roll E",
    "Ball Roll W",
    "Jr Nod"
  };


  // Here is where the user interface is created
  public void init() {
    int width = size().width;
    int height = size().height;
    button = new Button[bLabel.length];

    setLayout(new BorderLayout(10, 10));
    setBackground(new Color(128, 96, 128));

    viewerPanel = new Panel();
    luxoViewer = new LuxoCanvas(this, width - 10, height - 120);
    add(viewerPanel, BorderLayout.NORTH);
    viewerPanel.add(luxoViewer, null);

    buttonPanel = new Panel();
    add(buttonPanel, BorderLayout.SOUTH);
    buttonPanel.setLayout(new GridLayout(4, 4, 2, 2));
    for (int i = 0; i < bLabel.length; i++) {
      button[i] = new Button(bLabel[i]);
      button[i].setBackground(new Color(192, 160, 192));
      buttonPanel.add(button[i]);
    }

    luxoViewer.init();
  }

  // Here is where button presses are handled
  public boolean action(Event e, Object o) {
    if (e.target instanceof Button) {
      for (int i = 0; i < bLabel.length; i++) {
        if (bLabel[i].equals(o)) {
          luxoViewer.processButton(i);
          break;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  // makes a nice 5 pixel border around
  // the LuxoCanvas and buttonPanel
  public Insets insets() {
    return new Insets(5,5,5,5);
  }
}



class LuxoCanvas extends Canvas {
    final int CHUNKSIZE = 500;          // initial array size for vertices and triangles
    final int SKYCOLOR = 0xffd0e0ff;
    final int SKYCAM = 0;
    final int MOMCAM = 1;
    final int JRCAM = 2;
    
    
    Raster raster;
    Image screen;
    MatrixStack view;
    MatrixStack model;
    Vertex3D worldList[], screenList[];   // temporary storage for transformed vertices
    Luxo mom, jr;
    Vertex3D momEye, jrEye;
    Box floor;
    Ball ball;
    Light lightList[];
    int lights;
    int whichcam;
    
    Applet parent;

    // these variables are used to set up the viewing,
    // projection, and screen-space mapping matrices
    // several of these could have been Vertex3D Objects
    // but they really don't need some of the extra stuff
    // that a Vertex3D has, like normals and colors,
    // so I didn't decided to declare them as floats
    float eyex, eyey, eyez;
    float lookatx, lookaty, lookatz;
    float upx, upy, upz;
    float fov;
    int width, height;

    // i added these.
    float jrx, jry, jrbaseA, jrarmA, jrelbowA, jrneckA;
    float momx, momy, mombaseA, momarmA, momelbowA, momneckA;
    
    float ballx, bally, ballz, ballA;
    Vertex3D momLook, momUp, jrLook, jrUp;
    Vertex3D skyCam, skyLook, skyUp;
    Vertex3D skyXaxis, skynegZaxis, skyYaxis;
    Vertex3D defaultskycam, defaultskylook, defaultskyup, defaultskyxaxis, defaultskyyaxis, defaultskynegzaxis;
    Vertex3D momneck, momhead, jrneck, jrhead;
    
    float locallookx, locallooky, locallookz;

    // the constructor set the size of the rendering, creates
    // a raster, and fills it with the background color.
    public LuxoCanvas(Applet p, int w, int h) {
      super();
      parent = p;
      setSize(w, h);
      raster = new Raster(w, h);
      raster.fill(SKYCOLOR);
      screen = raster.toImage( );
      width = w;
      height = h;
    }

    // the init method resets the whole LuxoCanvas
    public void init( ) {
        // First, we create and initialize temporary storage for transformed
        // vertices. Here is a brief explanation of how these arrays get used.
        // Modeling coordinates of objects are transformed onto the worldList
        // where they are lit. These world coordinates then get transformed to
        // the screenList where they are rasterized. You can actually do this
        // trnsformation in one step, however, the method used here is more
        // flexible if you want to do some neat things later on (like shadow
        // maps).
        worldList = new Vertex3D[CHUNKSIZE];
        screenList = new Vertex3D[CHUNKSIZE];
        for (int i = 0; i < CHUNKSIZE; i++) {
            worldList[i] = new Vertex3D();
            screenList[i] = new Vertex3D();
        }

        // set default values for the viewing parameters
        eyex = 0;        eyey = -60;        eyez = 24;
        lookatx = 0;     lookaty = 0;     lookatz = 6;
        upx = 0;         upy = 0;         upz = 1;
        fov = 60;                           // in degrees

        // and my settings..
        jrx = 6.0f;
        jry = 0.0f;
        momx = -10.0f;
        momy = 0.0f;
        
        jrbaseA = toRadians(180f);
        jrarmA = toRadians(-135f);
        jrelbowA = toRadians(90f);
        jrneckA = toRadians(-60f);
        
        mombaseA = 0f;
        momarmA = toRadians(-120f);
        momelbowA = toRadians(120f);
        momneckA = toRadians(-60f);

        ballx= 3.0f;
        bally = 6.0f;
        ballz = 0.0f;
        ballA = 0.0f;

        whichcam = SKYCAM;

        // Set up two light sources, one ambient, one directional.
        // We'll get to this stuff later in class. Don't mess with
        // them at this point.
        lightList = new Light[10];
        lights = 0;
        lightList[lights++] = new Light(Light.AMBIENT, 0, 0, 0, 1.0f, 1.0f, 1.0f);
        lightList[lights++] = new Light(Light.DIRECTIONAL, 0.5f, 1.0f, -0.5f, 1.0f, 1.0f, 1.0f);

        // We next set up to matrix transform stacks.
        //
        // The view matrixStack will be used to map from world to screen
        // coordinates. It looks at the width and height of the Raster
        // to compute the mapping from canonical to screen coordinates
        view = new MatrixStack(raster);
        float t = (float) Math.sin(Math.PI*(fov/2)/180);
        float s = (t*height)/width;

        // Here we set the projection (from eye to canonical space)
        view.frustum(-t, t, -s, s, -1, -500);

        // Make a copy of the composed mapping from eye to screen space
        view.push();
        skyCam = new Vertex3D(eyex, eyey, eyez);
        skyLook = new Vertex3D(lookatx, lookaty, lookatz);
        skyUp = new Vertex3D(upx, upy, upz);  // already normalized
        
        //skyYaxis = skyUp;
        skynegZaxis = new Vertex3D((skyLook.x-skyCam.x),
                                 (skyLook.y-skyCam.y),
                                 (skyLook.z-skyCam.z));
                           
        normalize (skynegZaxis);
        
        skyXaxis = cross(skynegZaxis, skyUp); // ok normalized
        skyYaxis = cross(skyXaxis, skynegZaxis);  // ok normalized
        
       locallookx = dot(skyXaxis, skynegZaxis);
       locallooky = dot(skyYaxis, skynegZaxis);
       locallookz = -1 * dot(skynegZaxis, skynegZaxis);
                                                            
        
        // Set up a viewing matrix (hint: a camera)
       // view.lookAt(eyex, eyey, eyez, lookatx, lookaty, lookatz, upx, upy, upz);
       view.lookAt(skyCam.x, skyCam.y, skyCam.z, skyLook.x, skyLook.y, skyLook.z, skyUp.x, skyUp.y, skyUp.z);
       
            

        
        defaultskycam = new Vertex3D(eyex, eyey, eyez);
        defaultskylook = new Vertex3D(lookatx, lookaty, lookatz);
        defaultskyup = new Vertex3D(upx, upy, upz);
        defaultskynegzaxis = new Vertex3D(skynegZaxis.x, skynegZaxis.y, skynegZaxis.z);
        defaultskyxaxis = new Vertex3D(skyXaxis.x, skyXaxis.y, skyXaxis.z);
        defaultskyyaxis = new Vertex3D(skyYaxis.x, skyYaxis.y, skyYaxis.z);

        // The model matrixStack will be used to map a model's local
        // coordinates into world coordinates, there are a lot of different
        // "local" coordinate systems used, so this matrixStack gets a lot
        // of use
        model = new MatrixStack();

        // Next we create our four models.
        // The lamps each have four components, each with their own
        // vertex lists, triangle lists, and local coordinate systems.
        // The other two models, the floor and the ball have only one
        // component.
        mom = new Luxo(new Surface(0.4f, 0.4f, 1.0f, 0.5f, 0.5f, 0.0f, 100.0f));
        jr = new Luxo(new Surface(1.0f, 0.2f, 0.2f, 0.5f, 0.5f, 0.0f, 100.0f));
        floor = new Box(-24.0f, 24.0f, -24.0f, 24.0f, -1.0f, 0.0f,
                        new Surface(0.8f, 0.7f, 0.5f, 0.8f, 0.2f, 0.0f, 100.0f));
        ball = new Ball(new Surface(1.0f, 0.0f, 1.0f, 0.5f, 0.5f, 0.0f, 100.0f));

        // Reproportion Mom to make Jr model.
        // Here we use the model matrixStack to modify the vertex list
        // of Jr. There are other ways of doing this. For instance,
        // we could have declared just one lamp model and reproportion
        // and recolor it everytime we render. I've decided to keep things
        // conceptually, simple here. Each object has it's own model.
        // This code simply rescales the body and neck of the Jr model
        // so that it will have more childlike proportions.
        model.scale(0.5f, 1.0f, 1.0f);
        Vertex3D vList[] = jr.getVertices(Luxo.BODY);
        model.transform(vList, vList, vList.length);
        vList = jr.getVertices(Luxo.NECK);
        model.transform(vList, vList, vList.length);
        model.loadIdentity();

        // Now that all of the models are set up, call the
        // method that renders the whole scene.
        DrawScene();
        repaint();
        //view.pop();
    }
    
    public void normalize(Vertex3D vert) {
        float length = vert.x*vert.x + vert.y*vert.y + vert.z*vert.z;
        length = (float)Math.sqrt(length);
        vert.x = vert.x / length;
        vert.y = vert.y / length;
        vert.z = vert.z / length;
        
    }
    
    public float dot(Vertex3D v1, Vertex3D v2) {
        return (v1.x * v2.x + v1.y*v2.y + v1.z*v2.z);
    }
    
    public Vertex3D cross(Vertex3D v1, Vertex3D v2) {
        Vertex3D cp = new Vertex3D();
        cp.x = v1.y * v2.z - v1.z * v2.y;
        cp.y = v1.z * v2.x - v1.x * v2.z;
        cp.z = v1.x * v2.y - v1.y * v2.x;
        return cp;
    }
    

    public void paint(Graphics g) {
        g.drawImage(screen, 0, 0, this);
    }

    public void update(Graphics g) {
        paint(g);
    }

    float old_x, old_y;

    public boolean mouseDown(Event e, int x, int y) {
        // You will need to make extensive changes to this code
        // to implement the various camera navigation modes
        if (e.controlDown()) {
            parent.showStatus("Resetting model matrix");
            long time = System.currentTimeMillis();
            for (int i = 0; i <= 10; i++) {
              model.loadIdentity();
              // rotate the world around the up axis
              model.rotate(upx, upy, upz, (float)(i*Math.PI/5));
              DrawScene();
              update(this.getGraphics());
            }
            time = System.currentTimeMillis() - time;
            parent.showStatus("Time = "+(time/10)+" ms per frame");
        }
        old_x = x;
        old_y = y;
	      return true;
    }

    public boolean mouseDrag(Event e, int x, int y) {
        // You will need to make extensive changes to this code
        // to implement the various camera navigation modes
        if (e.controlDown()) {
            parent.showStatus("Resetting model matrix");
        } 
        
        else {
            if (e.metaDown()) {  // right button
                
                           
           switch (whichcam) {
                case SKYCAM:
           float dx = ((x-old_x)/20);
           float dy = ((old_y-y)/20);
           //dx = 0;
            old_y = x;
            old_y = y;
                
           if ((dx*dx) > (dy*dy)) dy = 0; else dx = 0;
                
           
           skyCam.x += (skyXaxis.x * dx + skyYaxis.x * dy);
           skyCam.y += (skyXaxis.y * dx + skyYaxis.y * dy);
           skyCam.z += (skyXaxis.z * dx + skyYaxis.z * dy);
           
           skyLook.x += (skyXaxis.x * dx + skyYaxis.x * dy);
           skyLook.y += (skyXaxis.y * dx + skyYaxis.y * dy);
           skyLook.z += (skyXaxis.z * dx + skyYaxis.z * dy);
           
           parent.showStatus("skycam moved "+(skyXaxis.x)+"x"+dx+"y"+skyXaxis.z * dx);
                   skynegZaxis = new Vertex3D((skyLook.x - skyUp.x * skyLook.x),
                                 (skyLook.y - skyUp.y * skyLook.y),
                                 (skyLook.z - skyUp.z * skyLook.z));
                                        
    
            DrawScene();
            repaint();
            
            break;
            
                case MOMCAM:
               float theta = (float) ((x - old_x)/20);
             float   theta2 = (float) ((y - old_y)/20);
                     if ((theta*theta) < (theta2*theta2)) theta = 0; else theta2 = 0;
            old_x = x;
            old_y = y;
            mombaseA -= theta;
            momelbowA +=theta2;
            CalcMom();
            //setMomCam();        
            DrawScene();
            repaint();
            //view.pop();
            parent.showStatus("momcam moved "+theta);
            break;
                
           
                case JRCAM:
                theta = (float) ((x - old_x)/20);
                theta2 = (float) ((y - old_y)/20);
                     if ((theta*theta) < (theta2*theta2)) theta = 0; else theta2 = 0;
                     
            old_x = x;
            old_y = y;
            jrbaseA -= theta;
            jrelbowA +=theta2;
            CalcJr();
            //setJrCam();        
            DrawScene();
            repaint();
            //view.pop();
            parent.showStatus("jrcam moved "+theta);
            break;
                
            }
            }
            else {  // normal left button
                
              switch (whichcam) {
                case SKYCAM:

                float theta = (float) toRadians(x - old_x);
                float theta2 = (float) toRadians(y - old_y);
                     if ((theta*theta) < (theta2*theta2)) theta = 0; else theta2 = 0;
            old_x = x;
            old_y = y;
            
            view.pop();
            view.push();
            view.loadIdentity();
            view.rotate(skyYaxis.x, skyYaxis.y, skyYaxis.z, theta);
            //view.rotate(skynegZaxis.x, skynegZaxis.y, skynegZaxis.z, theta2);
            
            view.rotate(skyXaxis.x, skyXaxis.y, skyXaxis.z, theta2);
            skyXaxis = view.transform(skyXaxis);
            skynegZaxis = view.transform(skynegZaxis);
            skyUp = view.transform(skyUp);
            skyYaxis = view.transform(skyYaxis);
            
            view.loadIdentity();
            
            view.translate(-1f *skyCam.x, -1f *skyCam.y, -1f*skyCam.z);
            view.rotate(skyYaxis.x, skyYaxis.y, skyYaxis.z, theta);

            view.rotate(skyXaxis.x, skyXaxis.y, skyXaxis.z, theta2);
            view.translate(skyCam.x, skyCam.y, skyCam.z);

            skyLook = view.transform(skyLook);

           parent.showStatus(
           "skycam moved "+theta2+
           //+theta+" "+skyXaxis.x+" "+skyXaxis.y+" "+skyXaxis.x+" "
           +skyLook.x+" "+skyLook.y+" "+skyLook.z+" "
           +skyUp.x+" "+skyUp.y+" "+skyUp.z);

            DrawScene();
            repaint();
            //view.push();
            //view.lookAt(eyex, eyey, eyez, lookatx, lookaty, lookatz, upx, upy, upz);
            
            //view.pop();
            
            break;
            
                case MOMCAM:
                theta = (float) toRadians(x - old_x);
                theta2 = (float) toRadians(y - old_y);
                     if ((theta*theta) < (theta2*theta2)) theta = 0; else theta2 = 0;
            old_x = x;
            old_y = y; 
            mombaseA -= theta;
            momneckA += theta2;

            CalcMom();
            //setMomCam();        
            DrawScene();
            repaint();
            //view.pop();
            parent.showStatus("momcam moved "+theta);
            break;
               
                
                case JRCAM:
                theta = (float) toRadians(x - old_x);
                theta2 = (float) toRadians(y - old_y);
                     if (theta < theta2) theta = 0; else theta2 = 0;
            jrbaseA -= theta;
            jrneckA += theta2;

            CalcJr();
            //setJrCam();        
            DrawScene();
            repaint();
            //view.pop();
            parent.showStatus("jrcam moved "+theta);
            break;
            
            }
            
        }}
        
        return true;
    }

    // The renderObject() method does all of the work of rendering objects.
    // You will probably not need to modify this method, though you may if
    // you wish.
    private void renderObject(MatrixStack m, Vertex3D vList[], Triangle tList[]) {
        int verts = vList.length;
        int tris = tList.length;
        // transform the model's coordinates to the world-coodinate system
        // using MatrixStack m. Set the triangle's vertex list to these
        // coordinates, then light the triangles.
        m.transform(vList, worldList, verts);
        Triangle.setVertexList(worldList);
        for (int i = 0; i < tris; i++) {
          tList[i].Illuminate(lightList, lights);
        }

        // Next we transform our world space coordinates to screen space.
        // The we tell the triangles to use these vertices to draw themselves.
        view.transform(worldList, screenList, verts);
        Triangle.setVertexList(screenList);
        for (int i = 0; i < tris; i++) {
          tList[i].Draw(raster);
        }
    }

    private float toRadians(float degrees) {
      return (float) (degrees/180.0f * Math.PI);
    }

    // I wrote this little helper method to aid in drawing the
    // Mom luxo lamp. It limit's the model's articulations to
    // the allowed degrees of freedom. If you don't find it
    // helpful, you do not have to use it.
    private Vertex3D DrawMom() {
        model.rotate(0.0f, 0.0f, 1.0f, mombaseA);
        renderObject(model, mom.getVertices(Luxo.BASE), mom.getTriangles(Luxo.BASE));
        model.translate(0.0f, 0.0f, 2.5f);
        model.rotate(0.0f, 1.0f, 0.0f, momarmA);
        renderObject(model, mom.getVertices(Luxo.BODY), mom.getTriangles(Luxo.BODY));
        model.translate(12.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, momelbowA);
        renderObject(model, mom.getVertices(Luxo.NECK), mom.getTriangles(Luxo.NECK));
        model.translate(12.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, momneckA);
        renderObject(model, mom.getVertices(Luxo.HEAD), mom.getTriangles(Luxo.HEAD));
        // Hum... Why do I continue transforming points
        // after I've already drawn everything?
        model.translate(1.0f, 0.0f, -1.0f);
        return model.transform(new Vertex3D(0f, 0f, 0f));
    }
    
        private void CalcMom() {
        model.push();
        model.loadIdentity();
        model.translate(momx, momy, 0f);
        model.rotate(0.0f, 0.0f, 1.0f, mombaseA);

        model.translate(0.0f, 0.0f, 2.5f);
        model.rotate(0.0f, 1.0f, 0.0f, momarmA);

        model.translate(12.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, momelbowA);
        momneck = model.transform(new Vertex3D(0f,0f,0f));
       
        model.translate(12.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, momneckA);
        momhead = model.transform(new Vertex3D(0f, 0f, 0f));
        
        // Hum... Why do I continue transforming points
        // after I've already drawn everything?
        model.translate(1.0f, 0.0f, -1.0f);
        momEye = model.transform(new Vertex3D(0f, 0f, 0f));
        momLook = model.transform(new Vertex3D(0f, 0f, -1f));
        momUp = model.transform(new Vertex3D(1f, 0f, 0f));
        
        model.pop();
    }

    // This helper method to aids in drawing the Jr. luxo lamp.
    // It limit's the model's articulations to the allowed
    // degrees of freedom. Just like above.
    private Vertex3D DrawJr() {
        model.scale(0.75f, 0.75f, 0.75f);
        model.rotate(0.0f, 0.0f, 1.0f, jrbaseA);
        renderObject(model, jr.getVertices(Luxo.BASE), jr.getTriangles(Luxo.BASE));
        model.translate(0.0f, 0.0f, 2.5f);
        model.rotate(0.0f, 1.0f, 0.0f, jrarmA);
        renderObject(model,  jr.getVertices(Luxo.BODY), jr.getTriangles(Luxo.BODY));
        model.translate(6.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, jrelbowA);
        renderObject(model, jr.getVertices(Luxo.NECK), jr.getTriangles(Luxo.NECK));
        model.translate(6.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, jrneckA);
        renderObject(model, jr.getVertices(Luxo.HEAD), jr.getTriangles(Luxo.HEAD));
        // Opps... I did it again, I played with your heart
        model.translate(1.0f, 0.0f, -1.0f);
        return model.transform(new Vertex3D(0f, 0f, 0f));
    }

    private void CalcJr() {
        model.push();
        model.loadIdentity();
        model.translate(jrx, jry, 0.0f);
        
        model.scale(0.75f, 0.75f, 0.75f);
        model.rotate(0.0f, 0.0f, 1.0f, jrbaseA);

        model.translate(0.0f, 0.0f, 2.5f);
        model.rotate(0.0f, 1.0f, 0.0f, jrarmA);

        model.translate(6.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, jrelbowA);
        jrneck = model.transform(new Vertex3D(0f, 0f, 0f));
        
        model.translate(6.0f, 0.0f, 0.0f);
        model.rotate(0.0f, 1.0f, 0.0f, jrneckA);
        jrhead = model.transform(new Vertex3D(0f, 0f, 0f));
        
        // Opps... I did it again, I played with your heart
        model.translate(1.0f, 0.0f, -1.0f);
        jrEye = model.transform(new Vertex3D(0f, 0f, 0f));
        jrLook = model.transform(new Vertex3D(0f, 0f, -1f));
        jrUp = model.transform(new Vertex3D(1f, 0f, 0f));
        model.pop();
    }

    private void jrtotarget(Vertex3D target) {
        CalcJr();
       Vertex3D targetvect = new Vertex3D(target.x - jrx, 
                                          target.y - jry, 0f);
       Vertex3D basevect = new Vertex3D(1f, 0f, 0f);
       normalize(targetvect);
       float cosangle = dot(targetvect, basevect);
       float firstangle = (float) Math.acos(cosangle);
       if (targetvect.y < 0) firstangle = -1f * firstangle;
       jrbaseA = firstangle;
       
       CalcJr();
       Vertex3D jrpart = jrEye;
       /*Vertex3D bodyvect = new Vertex3D(mompart.x - momneck.x,
                                 mompart.y - momneck.y,
                                 mompart.z - momneck.z);*/
       Vertex3D targvect = new Vertex3D(jrpart.x - target.x,
                                        jrpart.y - target.y,
                                        jrpart.z - target.z);
       Vertex3D targvectbase = new Vertex3D(jrpart.x - target.x,
                                        jrpart.y - target.y,
                                        0);
       //normalize(targvect);
       normalize(targvect);
       normalize(targvectbase);
       float someangle = (float) Math.acos(dot(targvectbase, targvect));
       if (jrpart.z < target.z) someangle = -1f * someangle; 
       float newangle = (float) (someangle - jrarmA - jrelbowA - Math.PI/2);
      
       jrneckA = newangle;
       CalcJr();
    }

    private void momtotarget(Vertex3D target) {
        CalcMom();
       Vertex3D targetvect = new Vertex3D(target.x - momx, 
                                          target.y - momy, 0f);
       Vertex3D basevect = new Vertex3D(1f, 0f, 0f);
       normalize(targetvect);
       float cosangle = dot(targetvect, basevect);
       float firstangle = (float) Math.acos(cosangle);
       if (targetvect.y < 0) firstangle = -1f * firstangle;
       mombaseA = firstangle;
       
       CalcMom();
       Vertex3D mompart = momEye;
       /*Vertex3D bodyvect = new Vertex3D(mompart.x - momneck.x,
                                 mompart.y - momneck.y,
                                 mompart.z - momneck.z);*/
       Vertex3D targvect = new Vertex3D(mompart.x - target.x,
                                        mompart.y - target.y,
                                        mompart.z - target.z);
       Vertex3D targvectbase = new Vertex3D(mompart.x - target.x,
                                        mompart.y - target.y,
                                        0);
       //normalize(targvect);
       normalize(targvect);
       normalize(targvectbase);
       float someangle = (float) Math.acos(dot(targvectbase, targvect));
       if (mompart.z < target.z) someangle = -1f * someangle; 
       float newangle = (float) (someangle - momarmA - momelbowA - Math.PI/2);
      
       momneckA = newangle;
       CalcMom();
    }

    private Vertex3D ballstar() {
        return new Vertex3D(ballx, bally, ballz + 4.0f);
    }

    // Okay, you'll need to completely rewrite this method.
    void DrawScene() {
                switch (whichcam) {
           case SKYCAM:
               setSkyCam();
               break;
            case JRCAM:
         
                setJrCam();
                break;
            case MOMCAM:
                setMomCam();
                break;
                }
          model.loadIdentity();
        // set the background color and reset the depth buffer
        raster.fill(SKYCOLOR);
        raster.resetz();

        // Draw the table, its frame is the same as the world frame.
        // In other words, it is already in world coordinates.
        renderObject(model, floor.getVertices(1), floor.getTriangles(1));

        // Save the world. Then add a ball.
        // The ball, has it's own coordinate frame, that, by default,
        // sits on the plane z = 0.
        model.push();        
        model.translate(ballx, bally, ballz+2f);
//        model.translate(0f, 0f, (float)(-1f * Math.cos(ballA) + 1) * 2f);
       // model.translate(0f,-2f,0f);
        model.rotate(0f, 1f, 0f, ballA);
        model.translate(0f,0f,-2f);
        renderObject(model, ball.getVertices(1), ball.getTriangles(1));
        model.pop();

        // Save the world. Then add Mom, whos frame also sits at z = 0.
        model.push();
        model.translate(momx, momy, 0.0f);
        momEye = DrawMom();
        momLook = model.transform(new Vertex3D(0f, 0f, -1f));
        momUp = model.transform(new Vertex3D(1f, 0f, 0f));
        model.pop();

        // Save the world. Then add Jr, who frame sits, you guessed it, at 0.
        model.push();
        model.translate(jrx, jry, 0.0f);
        jrEye = DrawJr();
        jrLook = model.transform(new Vertex3D(0f, 0f, -1f));
        jrUp = model.transform(new Vertex3D(1f, 0f, 0f));



        model.pop();

        // Make the raster into and Image.
        screen = raster.toImage( );
        
                

    }

    void setSkyCam() {
        view.pop();
        view.push();
          view.lookAt(skyCam.x, skyCam.y, skyCam.z, skyLook.x, skyLook.y, skyLook.z, skyUp.x, skyUp.y, skyUp.z);
        
        
    }

    void setMomCam() {
        view.pop();
        view.push();
        view.lookAt(momEye.x, momEye.y, momEye.z,
                    momLook.x, momLook.y, momLook.z,
                    momUp.x - momEye.x, momUp.y - momEye.y, momUp.z - momEye.z);
        
    }
    
   
    void setJrCam() {
        view.pop();
        view.push();
        view.lookAt(jrEye.x, jrEye.y, jrEye.z,
                    jrLook.x, jrLook.y, jrLook.z,
                    jrUp.x - jrEye.x, jrUp.y - jrEye.y, jrUp.z - jrEye.z);
        
    }
    

    void rollBall(int here) {
        float disttogo = here;
        float increm = disttogo / .2f;
        float incremA = here * .05f * (float) Math.PI;
        for (int k = 0; k< (here * increm); k++) {
            ballx += (here * .2);
            ballA += incremA;
            DrawScene();
            update(this.getGraphics());
        }
        //ballx = here;
    }
    void jrnod() {
        for (int i = 0; i< 5; i++) {
            jrneckA += .1f;
            DrawScene();
              update(this.getGraphics());
            parent.showStatus(" "+i);
        }
        for (int i=0; i < 5; i++){
        
            jrneckA += -.1f;
       DrawScene();
              update(this.getGraphics());
            parent.showStatus(i+" ");
        }
        }    
    
    
    void momnod() {
        for (int i = 0; i< 5; i++) {
            momneckA += .1f;
            DrawScene();
              update(this.getGraphics());
            parent.showStatus(" "+i);
        }
        for (int i=0; i < 5; i++){
        
            momneckA += -.1f;
       DrawScene();
              update(this.getGraphics());
            parent.showStatus(i+" ");
        }
        }    
    // Here's where you handle all of the button presses
    // from the button Panel
    public void processButton(int i) {
      switch (i) {
        case 0:                 // Reset
          model.loadIdentity();
            whichcam = SKYCAM;
        jrx = 6.0f;
        jry = 0.0f;
        momx = -10.0f;
        momy = 0.0f;
        
        jrbaseA = toRadians(180f);
        jrarmA = toRadians(-135f);
        jrelbowA = toRadians(90f);
        jrneckA = toRadians(-60f);
        
        mombaseA = 0f;
        momarmA = toRadians(-120f);
        momelbowA = toRadians(120f);
        momneckA = toRadians(-60f);

        ballx= 3.0f;
        bally = 6.0f;
        ballz = 0.0f;
        ballA = 0.0f;

         view.pop();
          view.push();
          skyCam = new Vertex3D(defaultskycam.x, defaultskycam.y, defaultskycam.z);
          skyUp = new Vertex3D(defaultskyup.x, defaultskyup.y, defaultskyup.z);
          skyLook = new Vertex3D(defaultskylook.x, defaultskylook.y, defaultskylook.z);
          skyXaxis = new Vertex3D(defaultskyxaxis.x, defaultskyxaxis.y, defaultskyxaxis.z);
          skyYaxis = new Vertex3D(defaultskyyaxis.x, defaultskyyaxis.y, defaultskyyaxis.z);
          skynegZaxis = new Vertex3D(defaultskynegzaxis.x, defaultskynegzaxis.y, defaultskynegzaxis.z);
          view.lookAt(skyCam.x, skyCam.y, skyCam.z, skyLook.x, skyLook.y, skyLook.z, skyUp.x, skyUp.y, skyUp.z);

          DrawScene();
          repaint();
          break;
          
        case 1:
          whichcam = SKYCAM;
         // setSkyCam();
          DrawScene();
          repaint();
          break;
          
        case 2:
          whichcam = MOMCAM;
          //setMomCam();
          DrawScene();
          repaint();
          break;
          
        case 3:
        
          whichcam = JRCAM;
          
          //setJrCam();
          DrawScene();
          repaint();
          break;
          
        case 4:               // mom@jr
        
          momtotarget(jrEye);
          momtotarget(jrEye);
          momtotarget(jrEye);
          momtotarget(jrEye);
          //momatjr();
          
          //CalcMom();
          //  parent.showStatus("ok the angle is "+mombaseA+"jrIx "+jrEye.x+" jrIy"+jrEye.y+" momx"+momx+" momy"+momy);
          
          DrawScene();
          repaint();
          break;

          case 5:
            momtotarget(ballstar());
            momtotarget(ballstar());
            momtotarget(ballstar());
            momtotarget(ballstar());
            DrawScene();
            repaint();
            break;
          
          case 6:
             jrtotarget(momEye);
             jrtotarget(momEye);
             jrtotarget(momEye);
             jrtotarget(momEye);
             DrawScene();
             repaint();
             break;
             
          case 7:
             jrtotarget(ballstar());
             jrtotarget(ballstar());
             jrtotarget(ballstar());
             jrtotarget(ballstar());
             
             DrawScene();
             repaint();
             break;

           
          case 8:
            for(int j =0;j<10;j++) {
            jry += .1f;
              update(this.getGraphics());
            }
            CalcJr();
            
          DrawScene();
          repaint();
          parent.showStatus("ok moved n");
            break;
            
            
        case 9:
            jry = jry - 1;
            CalcJr();
            
          DrawScene();
          repaint();
          
          parent.showStatus("ok moved s");
            break;
            
        case 10:
            jrx = jrx + 1;
            CalcJr();
            
          DrawScene();
          repaint();
            break;
            
        case 11:
            jrx = jrx - 1;
            CalcJr();
            
          DrawScene();
          repaint();
            break;

            case 12:
            momnod();
            break;
            
            case 13:
            rollBall(1);
            break;

           case 14:
            rollBall(-1);
            break;

        case 15:
        jrnod();
        break;

        default:
          parent.showStatus("Button "+i+" was pressed");
          break;
      }
    }
}