//*******************************************************
// haplotypes.c
//
#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "htsnps.h"
#include "snps.h"
#include "idpower.h"

int main(int argc, char *argv[]) {

  char *snpfilename = "";
  char *blockfilename = "";
  char *knownsnpfilename = "";
  int nextarg = 1; // if the argument has no .. arguments, then we can see if the next char is another argument.

  int c;

  int wrongargs = argc; // if == 1, we print usage.

  Known = 0;


  while (--argc > 0 && (*++argv)[0] == '-') {
    nextarg = 1;
    while ((c = *++argv[0]) && nextarg)
      switch (c) {

      case 's':
	snpfilename = (*++argv);
	//printf ("the snpfile specified is %s\n", snpfilename);
	nextarg = 0;
	argc--;
	break;

      case 'b':
	blockfilename = (*++argv);
	//printf ("the blockfile specified is %s\n", blockfilename);
	nextarg = 0;
	argc--;
	break;

      case 'k':
	knownsnpfilename = (*++argv);
	//	printf ("the known snps file specified is %s\n", knownsnpfilename);
	Known = 1;
	nextarg = 0;
	argc--;
	break;
	
      }
  }
  if (wrongargs == 1) { // if there are no arguments
    printf("Usage: haplotypes -s snpfile -b blockfile [-k knownsnpfile]\n");
    printf("\t-s snpfile: name of complete snp file in eb.out format\n");
    printf("\t-b blockfile: name of SORTED blockfile, no header, in block.dat format\n");
    printf("\t-k knownsnpfile: same as snpfile\n\n");
  }
  else { // what to do now.
   
    calcCoverage(snpfilename, blockfilename, knownsnpfilename);
  }

}

//*******************************************************
// calcCoverage
//
// calculates what percentage of the genome is covered, etc.
// but what it really does now it just run the haplotype identification
// algorithms
//
calcCoverage(char* snpfilename, char* blockfilename, char* knownsnpfilename) {
  int totalblocks, totalSNPs, totalKnownSNPs;
  int coveredSNPs;

  totalblocks = readblocksfromfile(blockfilename);
  totalSNPs = readSNPfromfile(snpfilename);
  if (Known)
    totalKnownSNPs = readKnownSNPfromfile(knownsnpfilename);

  //  coveredSNPs = countcoveredSNPs(totalSNPs, totalblocks);
  findhaplotypes(totalSNPs, totalblocks);
}


//*******************************************************
// countcoveredSNPs
//
// not used right now. but will tell you how many snps are
// covered by the blocks.
//
int countcoveredSNPs(int totalSNPs, int totalblocks) {
  int snpnum = 0;
  int blocknum = 0;
  int notcovered = 0, covered = 0, pos;
  int unknown = 0;

  while (snpnum < totalSNPs) {
    pos = getSNPpos(snpnum);
    if (blocknum < totalblocks) {
      if (pos < blockarray[blocknum][0]) {
	notcovered++;
	snpnum++;
	if (Known && ! snpKnown(pos))
	  unknown++;
      }
      else if (pos > blockarray[blocknum][1])
	blocknum++;
      else { covered++; snpnum++; }
    }
    else {
      notcovered++;
      snpnum++;
      if (Known && ! snpKnown(pos))
	unknown++;
    }
  }

  printf("%d %d %d %d\n",
	 totalSNPs, covered, notcovered, unknown);
  return covered;
}

//*******************************************************
// HAPLOTYPES
//
//

//*******************************************************
// gethaplotype
//
// given snp indices, and a chromosome index, returns the
// haplotype at that location as a char*.
//
char* gethaplotype(int first, int last, int index) {
  int i;
  char* temphap;

  temphap = (char * ) malloc ( (last - first + 2) * sizeof(char));
  for (i = first; i <= last; i++) {
    temphap[i-first] = allelearray[i][index];
  }
  temphap[i-first]='\0';
  return temphap;
}

//*******************************************************
// getknownhaplotype
//
// same as gethaplotype, but with the known snp array.
// does not check if you are in known mode..
//
char* getknownhaplotype(int first, int last, int index) {
  int i;
  char* temphap;

  temphap = (char * ) malloc ( (last - first + 2) * sizeof(char));
  for (i = first; i <= last; i++) {
    temphap[i-first] = knownallelearray[i][index];
  }
  temphap[i-first]='\0';
  return temphap;
}

//*******************************************************
// hapsequal 
//
// are these haplotypes equal?
// if i had been smart, i would have used a strlib function.
//
int hapsequal(char * hap, char * comparehap) {
  int i = 0;

  while (hap[i] != '\0') { // assume that both char*s are the same size
    if (hap[i] != comparehap[i])
      return 0;
    i++;
  }

  return 1;
}

//*******************************************************
// addorcount
//
// if hap is a new haplotype, add it to the array. otherwise add it
// to the tally of the matching haplotype.
//
int addorcount(char* hap, struct haplotype haparray[], int index) {
  int i, found = 0;

  for (i = 0; i < index; i++) {
    if (hapsequal(hap, haparray[i].hap)) {
      found = 1;
      (haparray[i].count)++;
      return index;
    }
  }
  if (! found) {
    haparray[index].hap = hap;
    haparray[index].count = 1;
  }
  return index+1;      

}

//*******************************************************
// sorthaplotypes
//
// sort the haplotypes by frequency.
//
int sorthaplotypes(struct haplotype haparray[], int hapindex) {
  int i, j;
  char* temphap;
  int tempcount;

  for (i = 0; i < hapindex - 1; i++) {
    for (j = i; j < hapindex; j++) {
      if (haparray[i].count < haparray[j].count) {
	temphap = haparray[i].hap;
	haparray[i].hap = haparray[j].hap;
	haparray[j].hap = temphap;

	tempcount = haparray[i].count;
	haparray[i].count = haparray[j].count;
	haparray[j].count = tempcount;
      }
    }
  }
}

//*******************************************************
// printhaplotypes
//
int printhaplotypes(struct haplotype haparray[], int hapindex) {

  int i;
  for (i = 0; i < hapindex; i++) {
    printf("%s  %d\n", haparray[i].hap, haparray[i].count);
  }
  //  printf("\n");
}

//*******************************************************
// countmajorhaps
//
// using cutoff, count how many haplotypes it takes to cover cutoff percent
// of the sample chromosomes.
//
int countmajorhaps(struct haplotype haparray[], int hapindex, int cutoff) {
  int i = 0, total = 0;

  while (i < hapindex && total < cutoff) {
    total += haparray[i].count;
    i++;
  }

  return i;
}

//*******************************************************
// countcommonhaps
//
// count how many haplotypes have a frequency > cutoff.
//
int countcommonhaps(struct haplotype haparray[], int hapindex, int cutoff) {
  int i = 0, tally = 0;

  for (i = 0; i < hapindex; i++) {
    if (haparray[i].count > cutoff)
      tally ++;
  }

  return tally;
}

//*******************************************************
// getcommonhaps
//
// returns the array of common haplotypes.
// useful when we see what happens when we start removing snps.
//
struct haplotype* getcommonhaps(struct haplotype haparray[], int hapindex, int cutoff) {
  int i = 0, index = 0;
  struct haplotype* commonhaparray = 
    (struct haplotype *) malloc (hapindex * sizeof(struct haplotype));

  for (i = 0; i < hapindex; i++) {
    if (haparray[i].count > cutoff) {
      // i think i just moved the pointer. careful not to edit this value later.
      commonhaparray[index] = haparray[i];
      index++;
    }
  }
  return commonhaparray;
}

//*******************************************************
// countpercentcommonhaps
//
// what percent of the chromosomes are covered by the common haplotypes?
//
int countpercentcommonhaps(struct haplotype haparray[], int hapindex, int cutoff) {
  int i = 0, tally = 0;

  for (i = 0; i < hapindex; i++) {
    if (haparray[i].count > cutoff)
      tally += haparray[i].count;
  }

  return tally;
}


//*******************************************************
// counthaplotypes
// 
// how many total haplotypes do we have?
// this looks at common haplotypes, haplotypes that cover a certain percent
// of the data, and prints out stats.
//
int counthaplotypes(int firstsnp, int lastsnp) {
  int hapindex = 0;
  int i, majhaps, commonhaps, percentcommon;
  int numchroms = 100; //stop gap solution
  char* temphap;
  struct haplotype haparray[MAXHAPS];
  struct haplotype *commonhaparray;
  int htsnps;
  double percenthapsknown;

  for (i = 0; i < numchroms; i++) {
    temphap = gethaplotype(firstsnp, lastsnp, i);
    hapindex = addorcount(temphap, haparray, hapindex);
  }

  //  printhaplotypes(haparray, hapindex);
  sorthaplotypes(haparray, hapindex);
  majhaps = countmajorhaps(haparray, hapindex, 80);
  commonhaps = countcommonhaps(haparray, hapindex, 5);
  percentcommon = countpercentcommonhaps(haparray, hapindex, 5);

  commonhaparray = getcommonhaps(haparray, hapindex ,5);
  htsnps = gethtSNPs(commonhaparray, commonhaps);

  percenthapsknown = (double) numberhapsknown(commonhaparray, commonhaps, getSNPpos(firstsnp)) / percentcommon;

  printf ("\n%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%.2f\t", getSNPpos(firstsnp),
	  getSNPpos(lastsnp), lastsnp - firstsnp + 1, hapindex, majhaps, commonhaps,
	  percentcommon, htsnps, percenthapsknown);


  //  printhaplotypes(haparray, hapindex);

  return hapindex;
}


//*******************************************************
// countknownhaplotypes
// 
// repeat above, but for the known snps.
//
// first and last are known indices
//
int countknownhaplotypes(int firstsnp, int lastsnp) {
  int hapindex = 0;
  int i, majhaps, commonhaps, percentcommon;
  int numchroms = 100; //stop gap solution
  char* temphap;
  struct haplotype haparray[MAXHAPS];
  int htsnps;
  struct haplotype *commonhaparray;

  for (i = 0; i < numchroms; i++) {
    temphap = getknownhaplotype(firstsnp, lastsnp, i);
    hapindex = addorcount(temphap, haparray, hapindex);
  }


  sorthaplotypes(haparray, hapindex);
  majhaps = countmajorhaps(haparray, hapindex, 80);
  commonhaps = countcommonhaps(haparray, hapindex, 5);
  percentcommon = countpercentcommonhaps(haparray, hapindex, 5);

  commonhaparray = getcommonhaps(haparray, hapindex ,5);
  htsnps = gethtSNPs(commonhaparray, commonhaps);


  printf ("%d\t%d\t%d\t%d\t%d\t%d\t",
	  lastsnp - firstsnp + 1, hapindex, majhaps, commonhaps, percentcommon, htsnps);

  //  printhaplotypes(haparray, hapindex);

  return hapindex;
}


//*******************************************************
// findhaplotypes
//
// the umbrella function, which calls all of those functions above.
// this also counts the number of snps inside and outside of blocks.
//
int findhaplotypes(int totalSNPs, int totalblocks) {
  int snpnum = 0;
  int blocknum = 0;
  int firstsnp, lastsnp;
  int hapcount, estsnpsneeded, totalestsnps = 0;
  int notinblocks = 0, unknownoutside = 0, unknowninside = 0;
  int knownsnpnum = 0, knownnotinblocks = 0, knownfirstsnp, knownlastsnp, knownhapcount;

  printf("SNP1\tSNP2\tN_SNPS\tAllHaps\tHap80\tHapC\tN_HapC\tNeedSNP\t%%HapsKn\t");
  if (Known)
    printf("K_SNPS\tK_Haps\tK_Hap80\tK_HapC\tN_KHapC\tNeedSNP\t");
  //  printf("htSNPs\n");

  for (blocknum = 0; blocknum < totalblocks; blocknum++) {
    while (getSNPpos(snpnum) != blockarray[blocknum][0]) {
      snpnum++;
      notinblocks++;
      if (Known && ! snpKnown(getSNPpos(snpnum)))
	unknownoutside++;
    }
    firstsnp = snpnum;
    while (getSNPpos(snpnum) != blockarray[blocknum][1]) {
      snpnum++;
      if (Known && ! snpKnown(getSNPpos(snpnum)))
	unknowninside++;
    }
    lastsnp = snpnum;

    hapcount = counthaplotypes(firstsnp, lastsnp);
    estsnpsneeded = (int) ceil(log(hapcount)/log(2));
    totalestsnps += estsnpsneeded;

    snpnum++;

    if (Known) {
      while (getKnownPosition(knownsnpnum) < blockarray[blocknum][0]) {
	knownsnpnum++;
	knownnotinblocks++;
      }
      knownfirstsnp = knownsnpnum;

      while (getKnownPosition(knownsnpnum) < blockarray[blocknum][1]) {
	knownsnpnum++;
      }
      knownlastsnp = knownsnpnum;
      
      // i had the following commented out, as i wasnted using the values.
      knownhapcount = countknownhaplotypes(knownfirstsnp, knownlastsnp);
      //knownestsnpsneeded = (int) ceil(log(hapcount)/log(2));
      //knowntotalestsnps += estsnpsneeded;
      //

      knownsnpnum++;
    }      

  }
    
  while (snpnum < totalSNPs) {
    notinblocks++;
    if (Known && ! snpKnown(getSNPpos(snpnum)))
      unknownoutside++;
    else knownnotinblocks++;
    snpnum++;
  }


  printf("\n");
  printf("notinblocks/known: %d\t", notinblocks);
  printf("%d\n", knownnotinblocks);

}
