//*******************************************************
// htsnps.c
//
// identifies the htsnps.
// this gives you a close to optimal selection.
// i didn't do the proof to figure out if it actually is optimal.

#include <stdio.h>
#include <math.h>
#include "defs.h"
#include "htsnps.h"

int printSNParray(int*, int);
int removeidenticalSNPs(int*, int, struct haplotype[], int);
int getSNPsinHap(char*);

//*******************************************************
// gethtSNPs
//
// here's the approach i took:
//
// 1. remove all uninformative SNPs
//    these are SNPs where they are the same value for all haplotypes.
//
// 2. remove all identical SNPs
//    these are SNPs which have the same pattern across all haplotypes.
//    we keep the first one one and remove the rest.
//
// 3. identify necessary SNPs
//    look at haplotypes which differ by only one SNP. if they only differ
//    at one point, you need to have that point otherwise you won't be able
//    to tell them apart.
//
// 4. check if we can distinguish all of our SNPs yet.
//
// 5. if not, take one of these necessary SNPs, and split the haplotypes
//    into two groups based on it.
//
// 6. take the first group. repeat from step 4.
//
// 7. can we distinguish all of our SNPs yet? if not, using the minimum
//    number of SNPs to distinguish the group, and begin to test all the
//    SNPs until we can distinguish all the haplotypes.
//
// 8. repeat for the second group.
//
//
// a SNParray is used to indicate the status of each SNP in the
// haplotype.
// 0 - this SNP will be ignored, because it is uninformative, or because
//     we are ignoring it during this round.
// 1 - this SNP has not been eliminated yet, we don't know anything about
//     whether it's important or not.. we want to change 1's to 0 and 2.
// 2 - necessary SNP.
//
int gethtSNPs(struct haplotype haparray[], int hapindex) {
  int totalSNPs, i;
  int* SNParray;
  int minSNPs;
  int count = 0;

  minSNPs = ceil(log(hapindex) / log(2));

  // get number of SNPs
  totalSNPs = getSNPsinHap(haparray[0].hap);
  SNParray = (int *) malloc (sizeof(int)*totalSNPs);
  for (i = 0; i < totalSNPs; i++)
    SNParray[i] = 1;

  removeuninfoSNPs(SNParray, totalSNPs, haparray, hapindex);
  removeidenticalSNPs(SNParray, totalSNPs, haparray, hapindex);
  identifynecessarySNPs(SNParray, totalSNPs, haparray, hapindex);
  
  // if you want to check your selection of htSNPs..
  //  printhapswithhtSNPs(SNParray, totalSNPs, haparray, hapindex);

  recursegethtSNPs(SNParray, totalSNPs, haparray, hapindex);
  //  printSNParray(SNParray, totalSNPs);
  //  printhapswithhtSNPs(SNParray, totalSNPs, haparray, hapindex);

  for ( i = 0; i < totalSNPs; i++)
    if (SNParray[i] == 2)
      count ++;

  return count;

  //  if (checkdone(SNParray, totalSNPs, minSNPs))
  //printf("DONE\n");
}

//*******************************************************
// recursegethtSNPs
// 
// see above for description of what this does.
//
int recursegethtSNPs(int SNParray[], int totalSNPs, struct haplotype haparray[], int hapindex) {
  int i = 0,j,hapcount = 0, found = 0;
  struct haplotype* temphaparray = (struct haplotype*) malloc (hapindex * sizeof(struct haplotype));

  if (hapindex > 1) {
    while (i < totalSNPs && !found) {
      if (SNParray[i] == 2) {
	found = 1;
	SNParray[i] = 0;
	
	// here we break the haplotypes into two groups, and recurse.
	// first for SNP = 1
	hapcount = 0;
	for (j = 0; j < hapindex; j++) {
	  if (haparray[j].hap[i] == '1') {
	    temphaparray[hapcount] = haparray[j];
	    hapcount++;
	  }
	}
	if (hapcount > 0)
	  recursegethtSNPs(SNParray, totalSNPs, temphaparray, hapcount);	

	// then for SNP = 0
	hapcount = 0;

	for (j = 0; j < hapindex; j++) {
	  if (haparray[j].hap[i] == '0') {
	    temphaparray[hapcount] = haparray[j];
	    hapcount++;
	  }
	}
	if (hapcount > 0)
	  recursegethtSNPs(SNParray, totalSNPs, temphaparray, hapcount);
	SNParray[i] = 2;
      }
      i++;
    }
    
    // 
    if (! found) {
      findminSNPs(SNParray, totalSNPs, haparray, hapindex);
      //      printf("after findminSNPs: ");
      // printSNParray(SNParray, totalSNPs);
      //      recursegethtSNPs(SNParray, totalSNPs, haparray, hapindex);
    }
  }
}

//*******************************************************
// findminSNPs
//
// when there are no more necessary SNPs to identify, find the minimum
// htSNPs that can distinguish the subgroup.
//
int findminSNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {
  int minSNPs, i = 0, j = 0, k = 0, leaveloop, trythis, done = 0;
  int count = 0, count2 = -1;
  int *testarray;

  // what's the minimum number of SNPs we need to distinguish this group?
  minSNPs = (int)ceil(log(hapindex) / log(2));

  // count how many we have to work with
  for (i = 0; i < totalSNPs; i++) {
    if (SNParray[i] == 1)
      count++;
  }

  if (count == minSNPs) {

    // if there's only that many SNPs left, you need all of them.
    // mark all with 2
    for (i = 0; i < totalSNPs; i++) {
      if (SNParray[i] == 1)
	SNParray[i] = 2;
    }
    // return - we're done!
    return 1;
  }

  // if we're not so lucky, we'll start trying different combinations.
  // test array holds the current indices of SNPs that we're checking.
  // we need to look at all combinations of SNPs from minSNPs to count.
  testarray = (int *) malloc(count * sizeof(int));
  for (i = 0; i < count; i++) {
    testarray[i] = -1;
  }

  // start with first minSNPs
  for (i = 0; i < minSNPs; i++) {
    testarray[i] = i;
  }

  // we'll mark the SNPs with a "3" to indicate that this is a SNP
  // we're testing.
  j = -1;
  k = 0;
  for (i = 0; (i < totalSNPs) && k < minSNPs; i++)
    if (SNParray[i] == 1) {
      j++;
      if (testarray[k] == j) {
	SNParray[i] = 3;
	k++;
      }
    }

  // are we done? test to see if this combination works.
  done = testSNPs(SNParray, totalSNPs, haparray, hapindex);

  // if it doesn't work..
  while (!done) {

    // clean up marks from last attempt
    for (j = 0; j < totalSNPs; j++) {
      if (SNParray[j] == 3)
	SNParray[j] = 1;
    }

    // pick a new set of SNPs
    leaveloop = 0;
    k = 1;
    while (k <= minSNPs && !leaveloop) {
      if (testarray[minSNPs - k] < count - 1) {
	testarray[minSNPs - k]++;
	for (j = minSNPs - k + 1; j < minSNPs; j++) {
	  testarray[j] = testarray[j-1]+1;
	}
	leaveloop = 1;
      }
      else {
	k++;
      }
    }

    // if we've exhausted all combinations, use n+1 instead of n SNPs
    if (k > minSNPs) {
      minSNPs++;
      for (j = 0; j < minSNPs; j++)
	testarray[j] = j;
    }

    // try again
    j = -1;
    k = 0;
    for (i = 0; (i < totalSNPs) && k < minSNPs; i++)
      if (SNParray[i] == 1) {
	j++;
	if (testarray[k] == j) {
	  SNParray[i] = 3;
	  k++;
	}
      }

    // are we done?
    done = testSNPs(SNParray, totalSNPs, haparray, hapindex);
    
  } 
  
  // when done, change 3's to 2's.
  for (j = 0; j < totalSNPs; j++) {
      if (SNParray[j] == 3)
	SNParray[j] = 2;
  }
  
}

//*******************************************************
// testSNPs
//
// see if these SNPs can distinguish all haplotypes.
//
int testSNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {
  int i = 0,j,hapcount = 0, found = 0;
  struct haplotype* temphaparray = (struct haplotype*) malloc (hapindex * sizeof(struct haplotype));
  int oneval = 1, zeroval = 1;

  if (hapindex == 1) return 1;

  if (hapindex > 1) {
    while (i < totalSNPs && !found) {
      if (SNParray[i] == 3) {
	found = 1;
	SNParray[i] = 0;
	
	// first for SNP = 1
	for (j = 0; j < hapindex; j++) {
	  if (haparray[j].hap[i] == '1') {
	    temphaparray[hapcount] = haparray[j];
	    hapcount++;
	  }
	}
	if (hapcount > 0)
	  oneval = testSNPs(SNParray, totalSNPs, temphaparray, hapcount);
	
	hapcount = 0;
	// then for SNP = 0
	for (j = 0; j < hapindex; j++) {
	  if (haparray[j].hap[i] == '0') {
	    temphaparray[hapcount] = haparray[j];
	    hapcount++;
	  }
	}
	if (hapcount > 0)
	  zeroval = testSNPs(SNParray, totalSNPs, temphaparray, hapcount);
	SNParray[i] = 3;
      }
      i++;
    }
    
    if (! found) {
      return 0;
    }
    return (zeroval && oneval);
  }
	    
}
//*******************************************************
// identifynecessarySNPs
//
// mark necessary SNPs with a "2"
//
int identifynecessarySNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {

  int i, j, k, count, snpindex;

  for (i = 0; i < hapindex - 1; i++) {
    for (j = i+1; j < hapindex; j++) {
      count = 0;
      k = 0;
      while (k < totalSNPs && count <= 1) {
	if (SNParray[k])
	  if (haparray[i].hap[k] != haparray[j].hap[k]) {
	    count ++;
	    snpindex = k;
	  }
	k++;
      }
      if (count == 1)
	SNParray[snpindex] = 2;
    }
  }
  
}

//*******************************************************
// checkdone
//
// are we completely done? this does a simple check, by seeing
// if the SNPs left equal the minimum number of SNPs. if so,
// we're done!
int checkdone(int SNParray[], int totalSNPs, int minSNPs) {
  int i, count =0;
  for (i = 0; i< totalSNPs; i++) {
    if (SNParray[i])
      count++;
  }
  if (count == minSNPs)
    return 1;

  count = 0;
  for (i = 0; i < totalSNPs; i++)
    count+= (SNParray[i] == 2 || SNParray[i] == 0);
  if (count == totalSNPs)
    return 1;

  return 0;
}

//*******************************************************
// printSNParray
//
int printSNParray(int* SNParray, int totalSNPs) {
  int i;
  for (i = 0; i< totalSNPs; i++) {
    if (SNParray[i] == 1)
      printf("%d ", i+1);
    else if (SNParray[i] == 2)
      printf("%d* ", i+1);      
  }
  printf("\n");
}

//*******************************************************
// printhapswithhtSNPs
//
// prints haplotypes, but only prints out the htSNPs identified so far.
// this allows you to check and see if the htSNPs can distinguish all of
// the haps.
//
int printhapswithhtSNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {
  int i, j;
  for (i = 0; i < hapindex;  i++) {
    for (j = 0; j < totalSNPs; j++) {
      if (SNParray[j] == 2)
	printf("%c",haparray[i].hap[j]);
      else printf(" ");
    }
    printf("\n");
  }

}

//*******************************************************
// removeuninfoSNPs
//
int removeuninfoSNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {
  int i, j, index;

  for (i = 0; i < totalSNPs-1; i++) {
    if (SNParray[i]) {
      j = 1;
      while ((j < hapindex) && (haparray[0].hap[i] == haparray[j].hap[i]))
	j++;
      if (j == hapindex) SNParray[i] = 0;
    }
  }
}

//*******************************************************
// removeidenticalSNPs
//
int removeidenticalSNPs(int* SNParray, int totalSNPs, struct haplotype haparray[], int hapindex) {
  int i, j, index;

  for (i = 0; i < totalSNPs-1; i++) {
    if (SNParray[i])
      for (j = i+1; j < totalSNPs; j++) {
	index = 0;
	if (SNParray[j]) {
	  if (haparray[index].hap[i] == haparray[index].hap[j])
	    while ((index < hapindex) && (haparray[index].hap[i] == haparray[index].hap[j]))
	      index++;
	  // need to do the same for exactly correlated the opposite way..
	  else
	    while ((index < hapindex) && (haparray[index].hap[i] != haparray[index].hap[j]))
	      index++;
	  if (index == hapindex)
	    SNParray[j] = 0;
	}
      }    
  }
}

//*******************************************************
// getSNPsinHap
// 
// i think this is just strlen.
//
int getSNPsinHap(char* hap) {  
  int i = 0;

  while (hap[i] != '\0')
    i++;

  return i;
}
