// idpower.c
//
// here we look at:
// what amount of the data can be identified with known SNPs?
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "snps.h"
#include "defs.h"

int hapswithknownSNPs(struct haplotype*, int, int*, struct haplotype*);
//char* gethapsubset(int[], char*);
int gethapsubset(int[], char*, char*);
int getknownsnparray(int startSNP, int snpnum, int *sarray);

//*******************************************************
// numberhapsknown
//
// what amount of the haplotype data can we identify with reduced SNP information?
// here we calculate the % of common haplotypes we can uniquely distinguish with
// known SNPs.
int numberhapsknown(struct haplotype* haparray, int hapindex, int startSNP) {
  
  struct haplotype * temphaparray;
  int snpnum;
  int temphapindex;
  int *sarray;
  int numberunique, i;

  snpnum = strlen(haparray[0].hap);

  temphaparray = (struct haplotype*) malloc ((hapindex+1) * sizeof (struct haplotype));

  sarray = (int *) malloc ((snpnum) * sizeof (int));
  if ((temphaparray == NULL) || (sarray == NULL)) {
    (void)fprintf(stderr, "ERROR: Malloc failed, numberhapsknown");
    (void)exit(EXIT_FAILURE);    /* or return EXIT_FAILURE; */
  }

  getknownsnparray(startSNP, snpnum, sarray);

  temphapindex = hapswithknownSNPs(haparray, hapindex, sarray, temphaparray);
  free(sarray);
  sarray = NULL;


  numberunique = getnumberuniquehaps(temphaparray, temphapindex);
  for (i = 0; i < hapindex; i++) {
    free(temphaparray[i].hap);
  }
  free(temphaparray);
  temphaparray = NULL;

  return numberunique;
}


//*******************************************************
// getnumberuniquehaps
//
// get the number of haplotypes that can be uniquely identified using
// knownSNPs
//
int getnumberuniquehaps(struct haplotype* haparray, int hapindex) {
  int i, j, k;
  int len, total = 0;
  int * killarray;

  len = strlen(haparray[0].hap);

  killarray = (int *) malloc(hapindex * sizeof (int));
  if (killarray == NULL) {
    (void)fprintf(stderr, "ERROR: Malloc failed, getnumberuniquehaps");
    (void)exit(EXIT_FAILURE);    /* or return EXIT_FAILURE; */
  }

  for ( i = 0; i < hapindex; i++)
    killarray[i] = 0;

  for ( i = 0; i < hapindex -1; i++) {
    for ( j = i + 1; j < hapindex; j++) {
      k = 0;
      while ((k < len) && (haparray[i].hap[k] == haparray[j].hap[k]))
	k++;
      if (k == len) {
	killarray[i] = 1;
	killarray[j] = 1;
      }
    }
  }

  for (i = 0; i < hapindex; i++) {
    if (killarray[i] == 0)
      total += haparray[i].count;
  }

  free(killarray);
  killarray = NULL;
  return total;
}

//*******************************************************
// getknownsnparray
//
// make an array, sarray, that indicates which of the SNPs within a
// given haplotype block is known. 0 - unknown, 1 - known.
//
int getknownsnparray(int startSNP, int length, int* sarray) {
  int i, index = 0, pos;

  index = getindexfrompos(startSNP);
  for ( i = 0 ; i < length; i ++)
    sarray[i] = 0;
  for ( i = 0; i < length; i++) {
    pos = getSNPpos(index + i);
    if (snpKnown(pos))
      sarray[i] = 1;
  }

  return i;
}

//*******************************************************
// hapswithknown
//
// fills temphaparray with the haplotypes using known SNPs
//
int hapswithknownSNPs(struct haplotype *haparray, int hapindex, int *sarray, struct haplotype *temphaparray) {
 
  //  char * temphap;
  int i, len, j;
  
  char * tempstr;

  len = strlen(haparray[0].hap);

  for (i = 0; i < hapindex; i++) {
    tempstr = (char*) malloc ((len+1) * sizeof(char));
    if (tempstr == NULL) {
      (void)fprintf(stderr, "ERROR: Malloc failed, hapswithknownSNPS");
      (void)exit(EXIT_FAILURE);    /* or return EXIT_FAILURE; */
    }
    //  temphaparray[i].hap = gethapsubset(sarray, haparray[i].hap, tempstr);
    gethapsubset(sarray, haparray[i].hap, tempstr);
    temphaparray[i].hap = tempstr;
    temphaparray[i].count = haparray[i].count;
    //    temphaparray[i].hap = temphap;
  }
  
  return i;
}

//*******************************************************
// gethapsubset
//
// get the haplotype using known SNPs
//
int gethapsubset(int* sarray1, char *hap1, char *ts) {
  int i = 0, index = 0;
  //  char * ts;
  int len;

  len = strlen(hap1);
 
  index = 0;
  
  for (i = 0; i < len; i++) {
    if (sarray1[i] == 1) {
      ts[index] = hap1[i];
      index++;
    }
  }

  ts[index] = '\0';
}

